// Pathfinder
// Writen by David J. Stein, Esq., June 2005. See "License.html" for terms and conditions of use.
//
// This module contains the core classes for the Pathfinder A* algorithm.

#region Using declarations

using System;

#endregion

namespace Pathfinder_NS {

	public enum Face {None, Left, Top, Right, Bottom};

	public class Pathfinder {

		#region Variables and initialization code

		public PathNode ActiveStart, ActiveCurrent, ActiveGoal;
		public bool bFoundSolution, bAnalysisComplete;
		public long lStepLength;
		public long lOffsetX, lOffsetY;
		public long lStartX, lStartY;
		public long lGoalX, lGoalY;
		public Location GotoLocation;
		public int iMajorDir, iMinorDir;
		public float fStepTally;
		public float fStepInterval;
		public CollisionRect StartR, StartB;

		public Pathfinder(long stx, long sty, long endx, long endy, CollisionRect str, CollisionRect stb, Location gotoloc, long sl) {
			lStepLength = sl;
			lStartX = stx;
			lStartY = sty;
			lOffsetX = lStartX % sl;
			lOffsetY = lStartY % sl;
			lGoalX = RoundXDown(endx);
			lGoalY = RoundYDown(endy);
			GotoLocation = gotoloc;
			ActiveStart = ActiveCurrent = ActiveGoal = null;
			bFoundSolution = false;
			StartR = str;
			StartB = stb;
			iMajorDir = iMinorDir = 0;
			fStepTally = fStepInterval = 0.0f;
			bFoundSolution = bAnalysisComplete = false;
		}

		#endregion

		#region Core functions

		public void SetupList() {
			// This function prepares the list for use.

			CollisionRect Search, Search2;
			long x, y, y2;
			PathNode NewNode, LastNode;

			ActiveStart = new PathNode();
			ActiveStart.iX = lStartX;
			ActiveStart.iY = lStartY;
			ActiveStart.lDistance = 0;
			ActiveStart.iBestDistanceToGoal = ActiveStart.iCombinedDistance = CalcDistance(lStartX, lStartY, lGoalX, lGoalY);
			ActiveStart.ActiveNext = null;
			LastNode = ActiveStart;
			Search = StartR;
			while (Search != null) {
				x = RoundXDown(Search.iLeft - 1);
				y = RoundYDown(Search.iTop - 1);
				if (((x != lStartX) || (y != lStartY)) && ((x != lGoalX) || (y != lGoalY))) {
					Search2 = StartR;
					while (Search2 != null) {
						if ((x < Search2.iLeft) || (x > Search2.iRight) || (y < Search2.iTop) || (y > Search2.iBottom))
							Search2 = Search2.NextR;
						else
							break;
					}
					if (Search2 == null) {  // add this node to the list
						NewNode = new PathNode();
						NewNode.iX = x;
						NewNode.iY = y;
						NewNode.iBestDistanceToGoal = CalcDistance(x, y, lGoalX, lGoalY);
						LastNode.ActiveNext = NewNode;
						LastNode = NewNode;
					}
				}
				y2 = RoundYUp(Search.iBottom + 1);
				if (((x != lStartX) || (y2 != lStartY)) && ((x != lGoalX) || (y2 != lGoalY))) {
					Search2 = StartR;
					while (Search2 != null) {
						if ((x < Search2.iLeft) || (x > Search2.iRight) || (y2 < Search2.iTop) || (y2 > Search2.iBottom))
							Search2 = Search2.NextR;
						else
							break;
					}
					if (Search2 == null) {  // add this node to the list
						NewNode = new PathNode();
						NewNode.iX = x;
						NewNode.iY = y2;
						NewNode.iBestDistanceToGoal = CalcDistance(x, y2, lGoalX, lGoalY);
						LastNode.ActiveNext = NewNode;
						LastNode = NewNode;
					}
				}
				x = RoundXUp(Search.iRight + 1);
				if (((x != lStartX) || (y != lStartY)) && ((x != lGoalX) || (y != lGoalY))) {
					Search2 = StartR;
					while (Search2 != null) {
						if ((x < Search2.iLeft) || (x > Search2.iRight) || (y < Search2.iTop) | (y > Search2.iBottom))
							Search2 = Search2.NextR;
						else
							break;
					}
					if (Search2 == null) {  // add this node to the list
						NewNode = new PathNode();
						NewNode.iX = x;
						NewNode.iY = y;
						NewNode.iBestDistanceToGoal = CalcDistance(x, y, lGoalX, lGoalY);
						LastNode.ActiveNext = NewNode;
						LastNode = NewNode;
					}
					if (((x != lStartX) || (y2 != lStartY)) && ((x != lGoalX) || (y2 != lGoalY))) {
						Search2 = StartR;
						while (Search2 != null) {
							if ((x < Search2.iLeft) || (x > Search2.iRight) || (y2 < Search2.iTop) || (y2 > Search2.iBottom))
								Search2 = Search2.NextR;
							else
								break;
						}
						if (Search2 == null) {
							NewNode = new PathNode();
							NewNode.iX = x;
							NewNode.iY = y2;
							NewNode.iBestDistanceToGoal = CalcDistance(x, y2, lGoalX, lGoalY);
							LastNode.ActiveNext = NewNode;
							LastNode = NewNode;
						}
					}
					Search = Search.NextR;
				}
				LastNode.ActiveNext = new PathNode();
				LastNode = LastNode.ActiveNext;
				ActiveGoal = LastNode;
				LastNode.iX = lGoalX;
				LastNode.iY = lGoalY;
				ActiveCurrent = ActiveStart;
			}
		}

		public bool Solve() {
			// This function moves one step toward a solution.

			PathNode SortNode, SortParent, SearchNode, SearchParent;
			long tempdistance, tempdistance2;

			if (ActiveCurrent != null) {
				if (ActiveCurrent == ActiveGoal) {
					bFoundSolution = true;
					bAnalysisComplete = true;
				}
				else if (ActiveCurrent.lDistance == 999999) {
					bFoundSolution = false;  // no solution exists.
					bAnalysisComplete = true;
				}
				else if (TestPath(ActiveCurrent.iX, ActiveCurrent.iY, lGoalX, lGoalY) == true) {
					tempdistance = ActiveCurrent.lDistance + ActiveCurrent.iBestDistanceToGoal;
					if (tempdistance < ActiveGoal.lDistance) {  // better path to parent
						ActiveGoal.lDistance = ActiveGoal.iCombinedDistance = tempdistance;
						ActiveGoal.Parent = ActiveCurrent;
						SearchParent = ActiveCurrent;
						while (SearchParent.ActiveNext != ActiveGoal)
							SearchParent = SearchParent.ActiveNext;
						if (tempdistance < SearchParent.iCombinedDistance) {  // move goal node
							SortParent = ActiveCurrent;
							SortNode = ActiveCurrent.ActiveNext;
							while (SortNode != null) {
								if (SortNode.iCombinedDistance > tempdistance) {  // move activegoal here
									SearchParent.ActiveNext = ActiveGoal.ActiveNext;
									ActiveGoal.ActiveNext = SortNode;
									SortParent.ActiveNext = ActiveGoal;
									break;
								}
								SortParent = SortNode;
								SortNode = SortNode.ActiveNext;
							}
						}
					}
					ActiveCurrent = ActiveCurrent.ActiveNext;
				} else {  // optimize this node against all others
					SearchNode = ActiveCurrent.ActiveNext;
					SearchParent = ActiveCurrent;
					while (SearchNode != null) {
						tempdistance = ActiveCurrent.lDistance + CalcDistance(SearchNode.iX, SearchNode.iY, ActiveCurrent.iX, ActiveCurrent.iY);
						tempdistance2 = tempdistance + SearchNode.iBestDistanceToGoal;
						if (tempdistance2 < SearchNode.iCombinedDistance) {
							if (TestPath(SearchNode.iX, SearchNode.iY, ActiveCurrent.iX, ActiveCurrent.iY) == true) {  // this path is clear
								SearchNode.Parent = ActiveCurrent;
								SearchNode.lDistance = tempdistance;
								SearchNode.iCombinedDistance = tempdistance2;
								// sort node into tree
								SortParent = ActiveCurrent;
								SortNode = ActiveCurrent.ActiveNext;
								while (SortNode != null) {
									if (SortNode.iCombinedDistance > tempdistance2) {
										if (SortParent != SearchNode) {  // move search_node here
											SearchParent.ActiveNext = SearchNode.ActiveNext;  // remove SearchNode from list
											SearchNode.ActiveNext = SortNode;
											SortParent.ActiveNext = SearchNode;
											SearchNode = SearchParent.ActiveNext;
											break;
										} else {  // we don't need to move it anywhere - exit the loop.
											SearchParent = SearchNode;
											SearchNode = SearchNode.ActiveNext;
											break;
										}
									} else {
										SortParent = SortNode;
										SortNode = SortNode.ActiveNext;
									}
								}
							} else {
								SearchParent = SearchNode;
								SearchNode = SearchNode.ActiveNext;
							} 
						}		 else {
							SearchParent = SearchNode;
							SearchNode = SearchNode.ActiveNext;
						}
					}
					ActiveCurrent = ActiveCurrent.ActiveNext;
				}
				if (ActiveCurrent == ActiveGoal) {
					bFoundSolution = true;
					bAnalysisComplete = true;
				}
				else if (ActiveCurrent.lDistance == 999999) {
					bFoundSolution = false;  // no solution exists.
					bAnalysisComplete = true;
				}
			}
			return bAnalysisComplete;
		}

		public int Step(long x, long y) {
			// This function takes the next step along a previously calculated path, based on run-slice computation and unrolled loops for improved performance.

			long dx, dy;
			if ((iMajorDir == 0) && (iMinorDir == 0) && (fStepInterval == 0.0f)) {  // set up paths originally
				PathNode current;
				if (bFoundSolution != false) {
					current = ActiveGoal;
					while (current.Parent != null) {
						current.Parent.Child = current;
						current = current.Parent;
					}
				}
				ActiveCurrent = ActiveStart;  // star tout at node right after ActiveNext
			}
			while ((ActiveCurrent.iX == x) && (ActiveCurrent.iY == y)) {  // done with this branch
				if (ActiveCurrent == ActiveGoal)
					return 255;  // at goal
				ActiveCurrent = ActiveCurrent.Child;
				fStepTally = 0.5f;
				if (ActiveCurrent.iX == x) {
					if (ActiveCurrent.iY > y) {  // always go down
						iMajorDir = 8;
						iMinorDir = 0;
						fStepInterval = 0;
					} else {  // always go up
						iMajorDir = 2;
						iMinorDir = 0;
						fStepInterval = 0;
					}
				}
				else if (ActiveCurrent.iY == y) {
					if (ActiveCurrent.iX > x) {  // always go right
						iMajorDir = 4;
						iMinorDir = 0;
						fStepInterval = 0;
					} else {  // always go left
						iMajorDir = 1;
						iMinorDir = 0;
						fStepInterval = 0;
					}
				}
				else {
					dx = (long) Math.Abs(ActiveCurrent.iX - x);
					dy = (long) Math.Abs(ActiveCurrent.iY - y);
					if (ActiveCurrent.iX > x) {
						if (ActiveCurrent.iY > y) {
							if (dx > dy) {
								iMajorDir = 4;
								iMinorDir = 12;
								fStepInterval = (float) (dy / dx);
							} else if (dx < dy) {
								iMajorDir = 8;
								iMinorDir = 12;
								fStepInterval = (float) (dx / dy);
							} else {  // always go down-right
								iMajorDir = 12;
								iMinorDir = 0;
								fStepInterval = 0;
							}
						} else {
							if (dx > dy) {
								iMajorDir = 4;
								iMinorDir = 5;
								fStepInterval = (float) (dy / dx);
							} else if (dx < dy) {
								iMajorDir = 2;
								iMinorDir = 6;
								fStepInterval = (float) (dx / dy);
							} else {  // always go up-right
								iMajorDir = 6;
								iMinorDir = 0;
								fStepInterval = 0;
							}
						}
					} else {
						if (ActiveCurrent.iY > y) {
							if (dx > dy) {
								iMajorDir = 1;
								iMinorDir = 0;
								fStepInterval = (float) (dy / dx);
							} else if (dx < dy) {
								iMajorDir = 8;
								iMinorDir = 9;
								fStepInterval = (float) (dx / dy);
							} else {  // always go down-left
								iMajorDir = 9;
								iMinorDir = 0;
								fStepInterval = 0;  // always go down-left
							}
						} else {
							if (dx > dy) {
								iMajorDir = 1;
								iMinorDir = 3;
								fStepInterval = (float) (dy / dx);
							} else if (dx < dy) {
								iMajorDir = 2;
								iMinorDir = 3;
								fStepInterval = (float) (dx / dy);
							} else {  // always go up-right
								iMajorDir = 3;
								iMinorDir = 0;
								fStepInterval = 0;
							}
						}
					}
				}
			}
			if (fStepInterval != 0) {
				fStepTally += fStepInterval;
				if (fStepTally > 1.0f) {
					fStepTally -= 1.0f;
					return iMinorDir;
				} else
					return iMajorDir;
			} else
				return iMajorDir;
		}

		#endregion

		#region Helper functions

		public bool TestPath(long x1, long y1, long x2, long y2) {
			// This function tells us whether the given path (x1,y1) - (x2,y2) collides with a rectangle. If so, it returns FALSE; if not, it returns TRUE. It's long and ugly because it uses unrolled loops, etc. for improved performance.

			CollisionRect ThisObstacle = null;
			long index;
			Face indexside;

			// page down to proper obstacle
			if (Math.Abs(x1 - x2) > Math.Abs(y1 - y2)) {  // X is a major axis - go by Y.
				index = Math.Min(y1, y2);
				ThisObstacle = StartB;
				while (ThisObstacle.iBottom < index) {
					ThisObstacle = ThisObstacle.NextB;
					if (ThisObstacle == null)  // all obstacles are above us - return true
						return true;
				}
				index = Math.Max(y1, y2);
				indexside = Face.Top;
			}
			else {  // Y is a major axis or line is slope (1) or (-1) - go by X.
				index = Math.Min(x1, x2);
				ThisObstacle = StartR;
				while (ThisObstacle.iRight < index) {
					ThisObstacle = ThisObstacle.NextR;
					if (ThisObstacle == null)  // all obstacles are to the left of us - return true
						return true;
				}
				index = Math.Max(x1, x2);
				indexside = Face.Left;
			}
			if (y1 < y2) {  // we're above our goal - check the top faces.
				if (x1 < x2) {  // we're to the left of our goal - check the left faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				} else if (x1 > x2) {  // we're to the right of our goal - check the right faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				} else {  // we're directly above our goal
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iTop))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				}
			} else if (y1 > y2) {  // we're below our goal - check the bottom faces.
				if (x1 < x2) {  // we're to the left of our goal - check the left faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				} else if (x1 > x2) {  // we're to the right of our goal - check the right faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
								else if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				} else {  // we're directly below our goal
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iBottom, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				}
			} else {  // we're on the same Y axis as our goal
				if (x1 < x2) {  // we're to the left of our goal - check the left faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iLeft, ThisObstacle.iTop, ThisObstacle.iLeft, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				} else if (x1 > x2) {  // we're to the right of our goal - check the right faces.
					if (indexside == Face.Left) {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextR;
						}
					} else {
						while (ThisObstacle != null) {
							if (ThisObstacle.Gateway != GotoLocation) {
								if (LinesIntersect(x1, y1, x2, y2, ThisObstacle.iRight, ThisObstacle.iTop, ThisObstacle.iRight, ThisObstacle.iBottom))
									return false;
							}
							ThisObstacle = ThisObstacle.NextB;
						}
					}
				}
			}
			return true;
		}

		public bool LinesIntersect(long x1, long y1, long x2, long y2, long x3, long y3, long x4, long y4) {
			long Ax, Bx, Cx, Ay, By, Cy, d, e, f;
			long x1lo, x1hi, y1lo, y1hi;

			Ax = x2 - x1;
			Bx = x3 - x4;
			if (Ax < 0) {  // X bound box test
				x1lo = x2;
				x1hi = x1;
			} else {
				x1hi = x2;
				x1lo = x1;
			}
			if (Bx > 0) {
				if ((x1hi < x4) || (x3 < x1lo))
					return false;
			} else {
				if ((x1hi < x3) || (x4 < x1lo))
					return false;
			}
			Ay = y2 - y1;
			By = y3 - y4;
			if (Ay < 0) {  // Y bound box test
				y1lo = y2;
				y1hi = y1;
			} else {
				y1hi = y2;
				y1lo = y1;
			}
			if (By > 0) {
				if ((y1hi < y4) || (y3 < y1lo))
					return false;
			} else {
				if ((y1hi < y3) || (y4 < y1lo))
					return false;
			}

			Cx = x1 - x3;
			Cy = y1 - y3;
			d = By * Cx - Bx * Cy;  // alpha numerator
			f = Ay * Bx - Ax * By;  // both denominator
			e = Ax * Cy - Ay * Cx;  // beta numerator
			if (f > 0) {  // alpha tests
				if ((d < 0) || (d > f))
					return false;
				else if ((e < 0) || (e > f))
					return false;
				else
					return true;
			}
			else {
				if ((d > 0) || (d < f))
					return false;
				else if ((e > 0) || (e < f))
					return false;
				else
					return true;
			}
		}

		public long CalcDistance(long x1, long y1, long x2, long y2) {
			return (long) Math.Sqrt(Math.Pow(x1 - x2, 2) + Math.Pow(y1 - y2, 2));
		}

		public long RoundXUp(long x) {
			while (x % lStepLength != lOffsetX)
				x++;
			return x;
		}

		public long RoundYUp(long y) {
			while (y % lStepLength != lOffsetY)
				y++;
			return y;
		}

		public long RoundXDown(long x) {
			while (x % lStepLength != lOffsetX)
				x--;
			return x;
		}
		
		public long RoundYDown(long y) {
			while (y % lStepLength != lOffsetY)
				y--;
			return y;
		}

		#endregion

	}

	public class PathNode {

		#region Variables and initialization code

		public long iX, iY, lDistance, iBestDistanceToGoal, iCombinedDistance;
		public PathNode ActiveNext, Child, Parent;

		public PathNode() {
			iX = 0;
			iY = 0;
			lDistance = 999999;
			iBestDistanceToGoal = 99999;
			iCombinedDistance = 999999;
			Parent = Child = null;
			ActiveNext = null;
		}

		#endregion

	}

	public class CollisionRect {

		#region Variables and initialization code

		public long iLeft, iTop, iRight, iBottom;
		public CollisionRect NextR, NextB;
		public Location Gateway;
		public long iGatewayDestX, iGatewayDestY;

		public CollisionRect() {
			NextR = NextB = null;
			Gateway = null;
			iGatewayDestX = iGatewayDestY = 0;
		}

		#endregion

	}

	public class CollisionRectList {

		#region Variables and initialization code

		public CollisionRect StartR, StartB;

		public CollisionRectList() {
			StartR = StartB = null;
		}

		#endregion

		#region Core functions

		public CollisionRect AddRect(long left, long top, long right, long bottom, Location Gateway, long gatewayx, long gatewayy) {
			// This function creates a new collisionrect for the given rect, location, and collision type.
			CollisionRect cl;
			CollisionRect insert;

			insert = new CollisionRect();
			insert.iLeft = left;
			insert.iTop = top;
			insert.iRight = right;
			insert.iBottom = bottom;
			insert.Gateway = Gateway;
			insert.iGatewayDestX = gatewayx;
			insert.iGatewayDestY = gatewayy;
			// sort by left side
			if (StartR == null)
				StartR = insert;
			else {
				if (StartR.iRight >= right) {  // insert first in list
					insert.NextR = StartR;
					StartR = insert;
				} else {
					cl = StartR;
					while ((cl.NextR != null) && (cl.NextR.iRight < right))
						cl = cl.NextR;
					insert.NextR = cl.NextR;
					cl.NextR = insert;
				}
			}
			// sort by bottom side
			// sort by left side
			if (StartB == null)
				StartB = insert;
			else {
				if (StartB.iBottom >= bottom) {  // insert first in list
					insert.NextB = StartB;
					StartB = insert;
				} else {
					cl = StartB;
					while ((cl.NextB != null) && (cl.NextB.iBottom < bottom))
						cl = cl.NextB;
					insert.NextB = cl.NextB;
					cl.NextB = insert;
				}
			}
			return insert;
		}

		#endregion

	}

	public class Location {
		// This class is a stub - carried over from legacy code.
	}
}
